package net.shambolica.helperl;

import com.ericsson.otp.erlang.*;
import net.shambolica.helperl.pattern.CompositePattern;
import net.shambolica.helperl.pattern.MatchContext;
import net.shambolica.helperl.pattern.OEValueBinder;
import net.shambolica.helperl.pattern.instance.*;

/**
 * Pattern matching.
 */
public class EP {


    /* =========================================================
     * General
     * ========================================================*/

    /**
     * Match any value.
     */
    public static CompositePattern pAny() {
        return new PAny();
    }


    /**
     * Same as {@link #pAny()}.
     */
    public static CompositePattern p_() {
        return pAny();
    }


    /**
     * Matches against the value.
     */
    public static CompositePattern pEq(final OtpErlangObject value) {
        return new PEq(value);
    }


    /**
     * Match any value, bind a key to it. If key already exists in result
     * with a <b>different</b> value, fail.
     */
    public static CompositePattern pSet(final String key) {
        return new PSet(key);
    }


    /**
     * Same as {@link #pSet(String)}.
     */
    public static CompositePattern p_(final String key) {
        return pSet(key);
    }


    /* =========================================================
     * Destructuring
     * ========================================================*/


    public static CompositePattern pTuple(final Object ... patterns) {
        return new PTuple(patterns);
    }


    public static CompositePattern pList(final Object ... patterns) {
        return new PList(patterns);
    }


    /* =========================================================
     * Expressions
     * ========================================================*/

    private static final PIsClass PATTERN_IS_ATOM = new PIsClass(
            OtpErlangAtom.class,
            OtpErlangBoolean.class);

    private static final PIsClass PATTERN_IS_BINARY = new PIsClass(
            OtpErlangBinary.class);

    private static final PIsClass PATTERN_IS_NUMBER = new PIsClass(
            OtpErlangByte.class,
            OtpErlangChar.class,
            OtpErlangDouble.class,
            OtpErlangFloat.class,
            OtpErlangInt.class,
            OtpErlangLong.class,
            OtpErlangShort.class,
            OtpErlangUInt.class,
            OtpErlangUShort.class);

    private static final PIsClass PATTERN_IS_INTEGER = new PIsClass(
            OtpErlangByte.class,
            OtpErlangChar.class,
            OtpErlangInt.class,
            OtpErlangLong.class,
            OtpErlangShort.class,
            OtpErlangUInt.class,
            OtpErlangUShort.class);

    private static final PIsClass PATTERN_IS_FLOAT = new PIsClass(
            OtpErlangDouble.class,
            OtpErlangFloat.class);

    private static final PIsClass PATTERN_IS_LIST = new PIsClass(
            OtpErlangList.class,
            OtpErlangString.class);

    private static final PIsClass PATTERN_IS_TUPLE = new PIsClass(
            OtpErlangTuple.class);

    private static final PIsClass PATTERN_IS_REF = new PIsClass(
            OtpErlangRef.class);

    private static final PIsClass PATTERN_IS_PID = new PIsClass(
            OtpErlangPid.class);

    private static final CompositePattern PATTERN_IS_TRUE = new CompositePattern() {
        @Override
        public boolean bind(OtpErlangObject target, MatchContext context) {
            return
                    ((target instanceof OtpErlangBoolean) && ((OtpErlangBoolean) target).booleanValue()) ||
                            ((target instanceof OtpErlangAtom) && ((OtpErlangAtom) target).atomValue().equals("true"));
        }
    };

    private static final CompositePattern PATTERN_IS_FALSE = new CompositePattern() {
        @Override
        public boolean bind(OtpErlangObject target, MatchContext context) {
            return
                    ((target instanceof OtpErlangBoolean) && !((OtpErlangBoolean) target).booleanValue()) ||
                            ((target instanceof OtpErlangAtom) && ((OtpErlangAtom) target).atomValue().equals("false"));
        }
    };

    private static final CompositePattern PATTERN_IS_BOOLEAN = new CompositePattern() {
        @Override
        public boolean bind(OtpErlangObject target, MatchContext context) {
            return
                    ((target instanceof OtpErlangBoolean) && ((OtpErlangBoolean) target).booleanValue()) ||
                            ((target instanceof OtpErlangAtom) && ((OtpErlangAtom) target).atomValue().equals("true")) ||
                            ((target instanceof OtpErlangBoolean) && !((OtpErlangBoolean) target).booleanValue()) ||
                            ((target instanceof OtpErlangAtom) && ((OtpErlangAtom) target).atomValue().equals("false"));
        }
    };


    /* Types */

    public static CompositePattern pIsAtom() {
        return PATTERN_IS_ATOM;
    }

    public static CompositePattern pIsAtom(String var) {
        return new PVarPredicate(pIsAtom(), var);
    }


    public static CompositePattern pIsBinary() {
        return PATTERN_IS_BINARY;
    }

    public static CompositePattern pIsBinary(String var) {
        return new PVarPredicate(pIsBinary(), var);
    }


    public static CompositePattern pIsNumber() {
        return PATTERN_IS_NUMBER;
    }

    public static CompositePattern pIsNumber(String var) {
        return new PVarPredicate(pIsNumber(), var);
    }


    public static CompositePattern pIsInteger() {
        return PATTERN_IS_INTEGER;
    }

    public static CompositePattern pIsInteger(String var) {
        return new PVarPredicate(pIsInteger(), var);
    }


    public static CompositePattern pIsFloat() {
        return PATTERN_IS_FLOAT;
    }

    public static CompositePattern pIsFloat(String var) {
        return new PVarPredicate(pIsFloat(), var);
    }


    public static CompositePattern pIsList() {
        return PATTERN_IS_LIST;
    }

    public static CompositePattern pIsList(String var) {
        return new PVarPredicate(pIsList(), var);
    }


    public static CompositePattern pIsTuple() {
        return PATTERN_IS_TUPLE;
    }

    public static CompositePattern pIsRef() {
        return PATTERN_IS_REF;
    }

    public static CompositePattern pIsPid() {
        return PATTERN_IS_PID;
    }


    /* Logic */

    public static CompositePattern pIsTrue() {
        return PATTERN_IS_TRUE;
    }

    public static CompositePattern pIsFalse() {
        return PATTERN_IS_FALSE;
    }

    public static CompositePattern pIsBoolean() {
        return PATTERN_IS_BOOLEAN;
    }


    public static CompositePattern pAnd(final OEValueBinder... patterns) {
        return new PAnd(patterns);
    }


    public static CompositePattern pOr(final OEValueBinder... patterns) {
        return new POr(patterns);
    }


    /* =========================================================
     * Internal
     * ========================================================*/


}
