package net.shambolica.helperl.pattern;

import com.ericsson.otp.erlang.OtpErlangObject;
import net.shambolica.helperl.util.Pair;

import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

import static net.shambolica.helperl.util.DataUtils.isEmpty;
import static net.shambolica.helperl.util.DataUtils.trim;

/**
 * A stack of variable bindings.
 */
public class MatchContext {

    private final Stack<Pair<String, OtpErlangObject>> bindings = new Stack<Pair<String, OtpErlangObject>>();


    /**
     * Stack management.
     */
    public int getMark() {
        return bindings.size();
    }


    /**
     * Stack management.
     */
    public void setMark(int mark) {
        if (bindings.size() < mark) {
            throw new IllegalStateException("mark out of depth");
        } else {
            if (mark < bindings.size()) {
                // Set popped values to null, to prevent leakage.
                for (int i = mark ; i < bindings.size() ; i++) {
                    bindings.set(i, null);
                }
            }
            bindings.setSize(mark);
        }
    }


    public boolean hasKey(String key) {
        key = normalizeKey(key);
        for (Pair<String, OtpErlangObject> pair : bindings) {
            if (pair.getA().equals(key)) {
                return true;
            }
        }
        return false;
    }


    public OtpErlangObject getBinding(String key) {
        key = normalizeKey(key);
        for (Pair<String, OtpErlangObject> pair : bindings) {
            if (pair.getA().equals(key)) {
                return pair.getB();
            }
        }
        return null;
    }


    /**
     * @return true if binding done or of same value as existing binding,
     * false if binding failed due to bad match.
     */
    public boolean bind(String key, OtpErlangObject value) {
        key = normalizeKey(key);
        if (value == null) {
            throw new IllegalArgumentException("null binding value");
        } else {
            OtpErlangObject binding = getBinding(key);
            if (binding == null) {
                bindings.add(new Pair<String, OtpErlangObject>(key, value));
                return true;
            } else {
                return binding.equals(value);
            }
        }
    }


    public Map<String,OtpErlangObject> getBindingMap() {
        Map<String,OtpErlangObject> map = new HashMap<String, OtpErlangObject>();
        for (Pair<String,OtpErlangObject> pair : bindings) {
            map.put(pair.getA(), pair.getB());
        }
        return map;
    }


    private String normalizeKey(String key) {
        key = trim(key);
        if (isEmpty(key)) {
            throw new IllegalArgumentException("empty key for match");
        } else {
            return key;
        }
    }

}
